# Copyright (C) 2025, Simona Dimitrova

import av
import av.container
import av.data.stream
import av.stream


class Stream():
    _stream: av.stream.Stream

    def __init__(self, stream: av.stream.Stream):
        self._stream = stream

    @property
    def type(self):
        return self._stream.type

    @property
    def index(self):
        return self._stream.index


class InputStream(Stream):
    pass


class OutputStream(Stream):
    def __init__(self, output_stream: av.stream.Stream, input_stream: InputStream = None):
        super().__init__(output_stream)
        self._input_stream = input_stream

    def process(self, packet):
        # Do nothing by default
        pass


class CopyOutputStream(OutputStream):
    def __init__(self, output_container: av.container.OutputContainer, input_stream: InputStream = None, encoder=None):
        if input_stream.type == "data":
            # DataStream.name is 'the codec'
            output_stream = output_container.add_data_stream(input_stream._stream.name)
        else:
            output_stream = output_container.add_stream_from_template(input_stream._stream)

        super().__init__(output_stream, input_stream)

    def process(self, packet):
        # Simply copy the packet, i.e. remux

        # We need to skip the "flushing" packets that `demux` generates.
        if packet.dts is None:
            return

        # We need to associate the packet with the this stream
        packet.stream = self
        self._stream.container.mux(packet._packet)



# Copyright (C) 2025, Simona Dimitrova

import av

from faceblur.av.stream import InputStream


class Packet():
    def __init__(self, packet: av.Packet, stream: InputStream):
        self._packet = packet
        self._stream = stream

    @property
    def dts(self):
        return self._packet.dts

    @property
    def pts(self):
        return self._packet.pts

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, stream):
        self._stream = stream
        self._packet.stream = stream._stream

    def decode(self):
        for frame in self._packet.decode():
            yield frame



# Copyright (C) 2025, Simona Dimitrova

import logging
import os
import tqdm

from faceblur.av.container import EXTENSIONS as CONTAINER_EXENTSIONS
from faceblur.av.container import FORMATS as CONTAINER_FORMATS
from faceblur.av.container import InputContainer, OutputContainer
from faceblur.av.video import THREAD_TYPE_DEFAULT
from faceblur.av.video import VideoFrame
from faceblur.faces.identify import identify_faces_from_image, identify_faces_from_video
from faceblur.faces.deidentify import blur_faces
from faceblur.image import EXTENSIONS as IMAGE_EXTENSIONS
from faceblur.image import FORMATS as IMAGE_FORMATS
from faceblur.threading import TerminatedException, TerminatingCookie

from PIL import Image
from pillow_heif import register_heif_opener
register_heif_opener()

DEFAULT_OUT = "_deident"

SUPPORTED_EXTENSIONS = set(CONTAINER_EXENTSIONS + IMAGE_EXTENSIONS)


def is_filename_from_ext_group(filename, group):
    _, ext = os.path.splitext(filename)
    return ext[1:].lower() in group


def _get_filenames_file(filename, on_error):
    if not is_filename_from_ext_group(filename, SUPPORTED_EXTENSIONS):
        on_error(f"Skipping unsupported file type: {os.path.basename(filename)}")
        return set()

    return set([filename])


def _get_filenames_dir(dirname, on_error):
    results = set()

    for root, dirs, files in os.walk(dirname, topdown=False):
        for name in files:
            results.update(_get_filenames_file(os.path.join(root, name), on_error))
        for name in dirs:
            results.update(_get_filenames_dir(os.path.join(root, name), on_error))

    return results


def get_supported_filenames(inputs, on_error=logging.getLogger(__name__).warning):
    filenames = set()

    for i in inputs:
        if os.path.isdir(i):
            filenames.update(_get_filenames_dir(i, on_error))
        elif os.path.isfile(i):
            filenames.update(_get_filenames_file(i, on_error))
        else:
            on_error(f"Invalid path: {i}")

    return set(sorted(list(filenames)))


def _create_output(filename, output, format=None):
    # Create the output directory
    os.makedirs(output, exist_ok=True)

    if format:
        is_image = is_filename_from_ext_group(filename, IMAGE_EXTENSIONS)
        formats = IMAGE_FORMATS if is_image else CONTAINER_FORMATS
        filename, ext = os.path.splitext(filename)
        ext = formats[format][0]
        filename = f"{filename}.{ext}"

    return os.path.join(output, os.path.basename(filename))


def _process_video_frame(frame: VideoFrame, faces, strength):
    # do extra processing only if any faces were found
    if faces:
        # av.video.frame.VideoFrame -> PIL.Image
        image = frame.to_image()

        # De-identify
        image = blur_faces(image, faces, strength)

        # PIL.Image -> av.video.frame.VideoFrame
        frame = VideoFrame.from_image(image, frame)

    return frame


def _faceblur_image(input_filename, output, strength, confidence, format):
    # Load
    image = Image.open(input_filename)

    # Find faces
    faces = identify_faces_from_image(image, detection_confidence=confidence)

    # De-identify
    image = blur_faces(image, faces, strength)

    # Save
    output_filename = _create_output(input_filename, output, format)
    image.save(output_filename)


def _faceblur_video(
        input_filename, output,
        strength, confidence,
        format, encoder,
        progress_type,
        thread_type, threads, stop):

    # First find the faces. We can't do that on a frame-by-frame basis as it requires
    # to have the full data to interpolate missing face locations
    with InputContainer(input_filename, thread_type, threads) as input_container:
        faces = identify_faces_from_video(
            input_container, detection_confidence=confidence, progress=progress_type, stop=stop)

    # let's reverse the lists so that we would be popping elements, rather than read + delete
    for frames in faces.values():
        frames.reverse()

    output_filename = _create_output(input_filename, output, format)
    try:
        with InputContainer(input_filename, thread_type, threads) as input_container:
            with OutputContainer(output_filename, input_container, encoder) as output_container:
                with progress_type(desc="Encoding", total=input_container.video.frames, unit=" frames", leave=False) as progress:
                    # Demux the packet from input
                    for packet in input_container.demux():
                        if packet.stream.type == "video":
                            for frame in packet.decode():
                                if stop:
                                    stop.throwIfTerminated()

                                # Get the list of faces for this stream and frame
                                faces_in_frame = faces[frame.stream.index].pop()

                                # Process (if necessary)
                                frame = _process_video_frame(frame, faces_in_frame, strength)

                                # Encode + mux
                                output_container.mux(frame)
                                progress.update()

                            if packet.dts is None:
                                # Flush encoder
                                output_container.mux(packet)
                        else:
                            # remux directly
                            output_container.mux(packet)
    finally:
        # Error/Stop request while encoding, make sure to remove the output
        try:
            os.remove(output_filename)
        except:
            pass


def faceblur(
        inputs,
        output,
        strength=1.0,
        confidence=0.5,
        video_format=None,
        video_encoder=None,
        image_format=None,
        progress_type=tqdm.tqdm,
        thread_type=THREAD_TYPE_DEFAULT,
        threads=os.cpu_count(),
        on_done=None,
        stop: TerminatingCookie = None):

    try:
        # Start processing them one by one
        with progress_type(get_supported_filenames(inputs), unit=" file(s)") as progress:
            for input_filename in progress:
                progress.set_description(desc=os.path.basename(input_filename))

                if stop:
                    stop.throwIfTerminated()

                if is_filename_from_ext_group(input_filename, IMAGE_EXTENSIONS):
                    # Handle images
                    _faceblur_image(input_filename, output, strength, confidence, image_format)
                else:
                    # Assume video
                    _faceblur_video(input_filename, output, strength, confidence, video_format,
                                    video_encoder, progress_type, thread_type, threads, stop)

                if on_done:
                    on_done(input_filename)
    except TerminatedException:
        # No action neccessary
        pass
    finally:
        if on_done:
            on_done(None)



# Copyright (C) 2025, Simona Dimitrova

import av.frame

from faceblur.av.stream import Stream


class Frame:
    def __init__(self, frame: av.frame.Frame, stream: Stream = None):
        self._frame = frame
        self._stream = stream

    @property
    def dts(self):
        return self._frame.dts

    @dts.setter
    def dts(self, dts):
        self._frame.dts = dts

    @property
    def pts(self):
        return self._frame.pts

    @pts.setter
    def pts(self, pts):
        self._frame.pts = pts

    @property
    def time_base(self):
        return self._frame.time_base

    @time_base.setter
    def time_base(self, time_base):
        self._frame.time_base = time_base

    @property
    def time(self):
        return self._frame.time

    @property
    def stream(self):
        return self._stream

    def copy_metadata(self, other):
        self._frame.dts = other._frame.dts
        self._frame.pts = other._frame.pts
        self._frame.time_base = other._frame.time_base
        self._stream = other._stream



# Copyright (C) 2025, Simona Dimitrova

import argparse
import av
import os

from faceblur.app import DEFAULT_OUT
from faceblur.app import faceblur
from faceblur.av.container import FORMATS as CONTAINER_FORMATS
from faceblur.av.video import ENCODERS, THREAD_TYPES, THREAD_TYPE_DEFAULT
from faceblur.image import FORMATS as IMAGE_FORMATS

av.logging.set_level(av.logging.ERROR)


def main():
    parser = argparse.ArgumentParser(
        description="A tool to obfuscate faces from photos and videos via blurring them."
    )

    parser.add_argument("inputs",
                        nargs="+",
                        help="Input file(s). May be photos or videos")

    parser.add_argument("--output", "-o",
                        default=DEFAULT_OUT,
                        help=f"Output folder for the blurred files. Defaults to {DEFAULT_OUT}.")

    parser.add_argument("--strength", "-s",
                        default=1.0, type=float,
                        help=f"""
                        Specify the strength of the deidentification.
                        It is a multiplier, so 0..1 makes them more recognisable,
                        while 1+ makes the less so.""")

    parser.add_argument("--video-format", "-f",
                        choices=sorted(list(CONTAINER_FORMATS.keys())),
                        help="""
                        Select a custom container format for video files.
                        If not speciefied it will use the same cotainer as each input.""")

    parser.add_argument("--video-encoder", "-v", choices=ENCODERS,
                        help="""
                        Select a custom video encoder.
                        If not speciefied it will use the same codecs as in the input videos""")

    parser.add_argument("--image-format", "-F",
                        choices=sorted(list(IMAGE_FORMATS.keys())),
                        help="""
                        Select a custom format for image files.
                        If not speciefied it will use the same format as each input.""")

    parser.add_argument("--thread_type", "-t",
                        choices=THREAD_TYPES,
                        default=THREAD_TYPE_DEFAULT,
                        help="PyAV decoder/encoder threading model")

    parser.add_argument("--threads", "-j",
                        default=os.cpu_count(), type=int,
                        help=f"""
                        How many threads to use for decoding/encoding video.
                        Defaults to the number of logical cores: {os.cpu_count()}.""")

    args = vars(parser.parse_args())
    faceblur(**args)


if __name__ == "__main__":
    main()



# Copyright (C) 2025, Simona Dimitrova

import av.stream

from faceblur.av.stream import InputStream


class Filter():
    def __init__(self, name, **params):
        self._name = name
        self._params = params

    @property
    def name(self):
        return self._name

    @property
    def params(self):
        return self._params


class Graph():
    _graph: av.filter.Graph

    def __init__(self, stream: InputStream, filters: list[Filter]):
        # Create the filter graph:
        # input buffer -> filter 1 -> filter 2 -> ... -> output buffersink
        graph = av.filter.Graph()

        # Prepare the input
        buffer = graph.add_buffer(template=stream._stream)

        # Input buffer is treated as the initial filter
        previous_filter = buffer

        for f in filters:
            # Convert to filter params:
            # param1=value1:param2=value2:...
            params = ":".join([f"{k}={v}" for k, v in f.params.items()]) if f.params else None

            # Previous filter -> filter
            filter = graph.add(f.name, params) if params else graph.add(f.name)
            previous_filter.link_to(filter)
            previous_filter = filter

        # Last filter -> output sink
        buffersink = graph.add("buffersink")
        previous_filter.link_to(buffersink)

        # Compile the graph
        graph.configure()
        self._graph = graph

    def push(self, frame: av.VideoFrame):
        self._graph.vpush(frame)

    def pull(self):
        return self._graph.vpull()



# Copyright (C) 2025, Simona Dimitrova

from PIL import ImageFilter
from PIL.Image import Image


MIN_FILTER_SIZE = 4
MAX_FILTER_SIZE = 1024
FACE_FILTER_DIVISOR = 20


def _calculate_filter_size(face, strength=1.0):
    return tuple(
        max(MIN_FILTER_SIZE, min(MAX_FILTER_SIZE, round(f / FACE_FILTER_DIVISOR) * strength))
        for f in (face.width, face.height)
    )


def blur_faces(image: Image, faces: list[tuple[int]], strength=1.0):
    for face in faces:
        # Crop the face region
        face_image = image.crop((face.left, face.top, face.right, face.bottom))

        # Calculate blur strength
        radius = _calculate_filter_size(face, strength)

        # Apply a Gaussian blur to the cropped region
        blurred_face_image = face_image.filter(ImageFilter.GaussianBlur(radius=radius))

        # Paste the blurred region back onto the image
        image.paste(blurred_face_image, (face.left, face.top, face.right, face.bottom))

    return image



# Copyright (C) 2025, Simona Dimitrova

FORMATS = {
    "bmp": ["bmp"],
    "png": ["png"],
    "jpeg": [
        "jpg",
        "jpe",
        "jpeg",
        "jfif",
    ],
    "jpeg2000": [
        "jp2",
        "j2k",
        "jpc",
        "jpf",
        "jpx",
        "j2c",
    ],
    "tiff": [
        "tif",
        "tiff",
    ],
    "webp": ["webp"],
    "heif": [  # pillow-heif
        "heic",
        "heics",
        "heif",
        "heifs",
        "hif",
    ],
    "tga": ["tga"],
    "ppm": [
        "ppm",
        "pbm",
        "pgm",
        "pnm",
        "pfm",
    ],
}

EXTENSIONS = sorted(list(set([ext for format in FORMATS.values() for ext in format])))



# Copyright (C) 2025, Simona Dimitrova

import av
import av.container
import av.stream
import logging
import pymediainfo

from faceblur.av.packet import Packet
from faceblur.av.stream import InputStream, OutputStream, CopyOutputStream
from faceblur.av.video import VideoPacket, VideoFrame, InputVideoStream, OutputVideoStream
from typing import Iterator

FORMATS = {
    "mjpeg": ["mjpg", "mjpeg"],                 # raw MJPEG video, Loki SDL MJPEG
    "wmv": ["wmv", "asf"],                      # ASF (Advanced / Active Streaming Format)
    "avi": ["avi"],                             # AVI (Audio Video Interleaved)
    "mpeg": ["mpg", "mpeg"],                    # MPEG-1 Systems / MPEG program stream
    "mpeg-ts": ["ts", "m2t", "mts", "m2ts"],    # MPEG-TS (MPEG-2 Transport Stream)
    "3gp": ["3gp"],                             # 3GP (3GPP file format)
    "3g2": ["3g2"],                             # 3GP2 (3GPP2 file format)
    "mp4": ["mp4"],                             # MP4 (MPEG-4 Part 14)
    "mov": ["mov"],                             # QuickTime / MOV
    "mkv": ["mkv"],                             # Matroska
    "webm": ["webm"],                           # WebM
    "raw.h261": ["h261"],                       # raw H.261
    "raw.h263": ["h263"],                       # raw H.263
    "raw.h264": ["h264"],                       # raw H.264 video
    "raw.hevc": ["hevc"],                       # raw HEVC video
    "raw.yuv": ["yuv"],                         # raw YUV video
    "raw.rgb": ["rgb"],                         # raw RGB video
}

EXTENSIONS = sorted(list(set([ext for format in FORMATS.values() for ext in format])))


class Container():
    def __init__(self, container: av.container.Container):
        self._container = container

    # Explicit close
    def close(self):
        """Close the container resource"""
        self._container.close()

    # Make sure not leaking on object destruction
    def __dealloc__(self):
        self.close()

    # Context manager (with/as)
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class InputContainer(Container):
    _container: av.container.InputContainer
    _video: InputVideoStream

    def __init__(self, filename: str, thread_type: str = None, thread_count: int = None):
        super().__init__(av.open(filename, metadata_errors="ignore"))

        self._info = pymediainfo.MediaInfo.parse(filename)
        self._duration = float(self._container.duration / av.time_base) if self._container.duration else 0

        # Update the thread type for the video decoders
        if thread_type is not None:
            for stream in self._container.streams.video:
                stream.thread_type = thread_type

        if thread_count is not None:
            for stream in self._container.streams.video:
                stream.thread_count = thread_count

        # Create dummy input streams for all non-video streams
        self._streams = {stream: InputStream(stream) for stream in self._container.streams if stream.type != "video"}

        # video stream infos (tracks in MediaInfo terms)
        tracks = self._info.video_tracks

        # If there is only one track and ID, the ID doesn't matter
        if (len(tracks) == 1) and (len(self._container.streams.video) == 1):
            stream = self._container.streams.video[0]
            self._streams[stream] = InputVideoStream(stream, vars(self._info.video_tracks[0]))
        else:
            # Multiple tracks require matching the track IDs
            # Reshape the tracks into a {id: track}
            tracks = {t.track_id: t for t in tracks}

            # Directly use the ID for container formats that support IDs, e.g. MOV, MPEG, etc., see AVFMT_SHOW_IDS.
            # If IDs are not supported, assume the ID from the index the way MediaInfo expects them to be
            self._streams.update({
                stream:
                InputVideoStream(
                    stream, vars(tracks[stream.id if self._container.format.show_ids else stream.index + 1]))
                for stream in self._container.streams.video})

        self._video = self._streams[self._container.streams.video[0]]

    @property
    def video(self):
        return self._video

    @property
    def streams(self):
        return tuple(self._streams.values())

    def demux(self) -> Iterator[Packet | VideoPacket]:
        for packet in self._container.demux():
            if packet.stream.type == "video":
                yield VideoPacket(packet, self._streams[packet.stream])
            else:
                yield Packet(packet, self._streams[packet.stream])


class OutputContainer(Container):
    _container: av.container.OutputContainer
    _streams: dict[InputStream, OutputStream]

    def __init__(self, filename: str, template: InputContainer = None, encoder=None):
        super().__init__(av.open(filename, "w"))

        self._streams = {}

        if template:
            # Create output streams matching the input ones
            for stream in template._streams.values():
                self.add_stream_from_template(stream, encoder)

    def add_stream_from_template(self, template: InputStream, encoder=None):
        STREAM_TYPES = {
            "audio": CopyOutputStream,
            "video": OutputVideoStream,
            # currently subtitles streams are not remuxed, as this needs to be tested
            # currently data streams are not remuxed, as no data encoders are present,
            # and creating a data stream without a codec only appears to work for .ts
        }

        if template.type not in STREAM_TYPES:
            # Don't handle unsupported stream types
            logging.getLogger(__name__).debug("Skipping unsupported stream type %s", template.type)
            return None

        # create the stream wrapper
        stream = STREAM_TYPES[template.type](self._container, template, encoder)

        # add to mappings of input -> output streams
        self._streams[template] = stream

        # and return to user
        return stream

    @property
    def streams(self):
        return tuple(self._streams.values())

    def mux(self, packet_or_frame: Packet | VideoFrame):
        if packet_or_frame.stream in self._streams:
            self._streams[packet_or_frame.stream].process(packet_or_frame)



# Copyright (C) 2025, Simona Dimitrova

import threading

from faceblur.exception import FaceblurException


class TerminatedException(FaceblurException):
    pass


class TerminatingCookie:
    def __init__(self):
        self._bomb = threading.Event()

    def requestTermination(self):
        self._bomb.set()

    def throwIfTerminated(self):
        if self._bomb.is_set():
            raise TerminatedException()



# Copyright (C) 2025, Simona Dimitrova

import av
import av.stream
import math
import pymediainfo

from faceblur.av.stream import InputStream, OutputStream
from faceblur.av.filter import Filter, Graph
from faceblur.av.frame import Frame
from faceblur.av.packet import Packet
from PIL.Image import Image

THREAD_TYPES = [
    "SLICE",
    "FRAME",
    "AUTO,"
]

THREAD_TYPE_DEFAULT = "AUTO"


def __check_codec(codec):
    try:
        codec = av.codec.Codec(codec, "w")
        return codec.type == "video"
    except av.codec.codec.UnknownCodecError:
        # Not encoder
        return False


ENCODERS = sorted([codec for codec in av.codecs_available if __check_codec(codec)])


def _get_angle360(angle: float):
    # Make sure the angle is in 0..360
    # See get_rotation() in fftools/cmdutils.c
    return angle - 360 * int(angle/360 + 0.9/360)


def _dimensions_for_rotated(width, height, angle):
    if abs(angle - 90) < 1:
        # Rotated by 90
        return height, width
    elif abs(angle - 270) < 1:
        # Rotated by 270
        return height, width
    else:
        return width, height


def _filters_for_rotated(angle, input_stream: InputStream):
    # Handle different rotations.
    # See `if (autorotate) {` in fftools/ffplay.c
    if abs(angle - 90) < 1:
        # Rotated by 90. Simple transpose is enough
        return [
            Filter("transpose", dir="clock"),
        ]
    elif abs(angle - 180) < 1:
        # Rotated by 180. Use horizontal and vertical flip
        return [
            Filter("hflip"),
            Filter("vflip"),
        ]
    elif abs(angle - 270) < 1:
        # Rotated by 270. Simple transpose is enough
        return [
            Filter("transpose", dir="cclock"),
        ]
    elif abs(angle) > 1:
        # Generic rotation by an odd angle
        return [
            Filter("rotate", angle=angle * (math.pi / 180)),
        ]
    else:
        # No rotation necessary (0 <= angle < 1)
        return None


class InputVideoStream(InputStream):
    _info: pymediainfo.Track
    _graph = None

    def __init__(self, stream: av.stream.Stream, info: pymediainfo.Track):
        super().__init__(stream)
        self._info = info

        # Get rotation from stream side data and fix the resolutions
        rotation = float(info.get("rotation", 0))
        cc = stream.codec_context
        angle = _get_angle360(rotation)
        self._width, self._height = _dimensions_for_rotated(cc.width, cc.height, angle)

        if rotation:
            # Need to create rotation filters
            filters = _filters_for_rotated(angle, stream)
            if filters:
                self._graph = Graph(self, filters)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def info(self):
        return self._info

    @property
    def frames(self):
        return self._stream.frames


class VideoFrame(Frame):
    def to_image(self) -> Image:
        return self._frame.to_image()

    @staticmethod
    def from_image(image: Image, frame: Frame):
        new_frame = av.VideoFrame.from_image(image)
        new_frame = Frame(new_frame)
        new_frame.copy_metadata(frame)
        return new_frame


class VideoPacket(Packet):
    _stream: InputVideoStream

    def decode(self):
        for frame in self._packet.decode():
            if self._stream._graph:
                self._stream._graph.push(frame)
                frame = self._stream._graph.pull()
            yield VideoFrame(frame, self.stream)


class OutputVideoStream(OutputStream):
    def __init__(self,
                 output_container: av.container.OutputContainer,
                 input_stream: InputVideoStream = None,
                 encoder: str = None):

        if not encoder:
            # Use same encoder as decoder
            encoder = input_stream._stream.codec.name

        # We need a concrete frame rate to pass to add_stream
        frame_rate = input_stream._stream.codec_context.framerate
        if not frame_rate:
            # variable frame rate, but some encoders don't seem to work fine with it
            # so use the guessed one
            frame_rate = round(input_stream._stream.guessed_rate)

        output_stream = output_container.add_stream(encoder, frame_rate)

        # Those parameters are from FFMPEG's avcodec_parameters_to_context(), which is
        # called from av.container.output.OutputContainer.add_stream_from_template().
        params = [
            # General
            "bit_rate",
            # "bits_per_coded_sample", # Not supported for encoders
            # "bits_per_raw_sample", # N/A
            "profile",
            # "level", # N/A

            # Video
            "pix_fmt",
            # do not read width/height directly from the codec context
            # "field_order", # N/A

            "color_range",
            "color_primaries",
            "color_trc",
            "colorspace",

            # "chroma_sample_location", # N/A
            "sample_aspect_ratio",
            # "has_b_frames", # Read-only
            "framerate",

            "extradata",

            # Copy over the thread config
            "thread_type",
            "thread_count",
        ]

        for p in params:
            value = getattr(input_stream._stream.codec_context, p)
            if value is not None:
                setattr(output_stream.codec_context, p, value)

        # Get the dimensions from the InputStream not from CodecContext,
        # in order to take any rotation into account
        output_stream.codec_context.width = input_stream.width
        output_stream.codec_context.height = input_stream.height

        super().__init__(output_stream, input_stream)

    def process(self, frame: VideoFrame):
        if frame.dts is not None:
            # Encode
            for packet_output in self._stream.encode(frame._frame):
                self._stream.container.mux(packet_output)
        else:
            # Flush the encoder.
            # Note that the "empty" packet MUST first be passed to the
            # encoder to signal flushing
            while True:
                try:
                    for packet_output in self._stream.encode(None):
                        self._stream.container.mux(packet_output)
                except av.error.EOFError:
                    break



# Copyright (C) 2025, Simona Dimitrova

import math
import numpy as np
import os
import tqdm

from faceblur.av.container import InputContainer
from faceblur.threading import TerminatingCookie
from mediapipe.python.solutions.face_detection import FaceDetection
from PIL.Image import Image


IDENTIFY_IMAGE_SIZE = 1920


def _find_divisor(width, height, max_side):
    side = max(width, height)
    return math.ceil(side / max_side)


def _merge_faces(faces_close, faces_far, confidence):
    faces = []

    while faces_close:
        faces_with_best_score = -1, -1
        best_score = 0

        # Find the best matches for each face
        for face_close_index, face_close in enumerate(faces_close):
            for face_far_index, face_far in enumerate(faces_far):
                score = face_close.intersection_over_union(face_far)
                if score > best_score:
                    faces_with_best_score = face_close_index, face_far_index
                    best_score = score

        if best_score >= confidence:
            # Found faces with enough similarity
            face_close_index, face_far_index = faces_with_best_score
            face_close = faces_close[face_close_index]
            face_far = faces_far[face_far_index]

            # Remove pair from further processing
            faces_close.remove(face_close)
            faces_far.remove(face_far)

            # And add the union of both into the faces found so far
            faces.append(face_close.union(face_far))
        else:
            # Could not find close faces that intersect with enough area
            # Therefore all left faces are distinct
            faces.extend(faces_close)
            break

    # If there are any left far faces, it means they must be distinct
    faces.extend(faces_far)

    return faces


def _identify_faces_from_image_array(image, width, height, face_detection):
    faces = []
    results = face_detection.process(image)
    if results.detections:
        max_box = Box(0, width - 1, height - 1, 0)

        for detection in results.detections:
            box = detection.location_data.relative_bounding_box

            # Adjust the faces as mediapipe returns relative data
            left = int(box.xmin * width)
            top = int(box.ymin * height)
            right = int((box.xmin + box.width) * width) - 1
            bottom = int((box.ymin + box.height) * height) - 1

            # Make sure the face box is within the image as detection may return coords out of bounds
            face = Box(top, right, bottom, left).intersect(max_box)
            faces.append(face)

    return faces


def _identify_faces_from_image(image: Image,
                               face_detection_close,
                               face_detection_far,
                               merge_confidence,
                               image_size):

    # Cache original dimension as results are normalised
    width = image.width
    height = image.height

    divisor = _find_divisor(image.width, image.height, image_size)
    if divisor > 1:
        # Needs to be scaled down
        image = image.resize((image.width // divisor, image.height // divisor))

    image = np.array(image)
    faces_close = _identify_faces_from_image_array(image, width, height, face_detection_close)
    faces_far = _identify_faces_from_image_array(image, width, height, face_detection_far)

    # If either model did not find faces, simply return whatever the other one found (if any)
    if not faces_close:
        return faces_far
    elif not faces_far:
        return faces_close

    # This means that both found faces and we need to group then and unite them into bigger boxes
    return _merge_faces(faces_close, faces_far, merge_confidence)


class Box:
    def __init__(self, top, right, bottom, left):
        if left > right:
            raise ValueError(f"left={left} > right={right}")

        # The coordinate are inverted on Y
        if top > bottom:
            raise ValueError(f"top={top} > bottom={bottom}")

        self.top = top
        self.right = right
        self.bottom = bottom
        self.left = left

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    def intersect(self, other):
        # Calculate the intersection coordinates
        intersection_top = max(self.top, other.top)
        intersection_right = min(self.right, other.right)
        intersection_bottom = min(self.bottom, other.bottom)
        intersection_left = max(self.left, other.left)

        # Check if there is an intersection
        if intersection_top <= intersection_bottom and intersection_left <= intersection_right:
            return Box(intersection_top, intersection_right, intersection_bottom, intersection_left)
        else:
            # No intersection
            return None

    def union(self, other):
        # Calculate the union coordinates
        union_top = min(self.top, other.top)
        union_right = max(self.right, other.right)
        union_bottom = max(self.bottom, other.bottom)
        union_left = min(self.left, other.left)
        return Box(union_top, union_right, union_bottom, union_left)

    def area(self):
        return (self.bottom - self.top + 1) * (self.right - self.left + 1)

    def __repr__(self):
        return f"Box(top={self.top}, right={self.right}, bottom={self.bottom}, left={self.left})"

    def __eq__(self, other):
        return self.top == other.top and self.right == other.right and self.bottom == other.bottom and self.left == other.left

    def intersection_over_union(self, other):
        intersection = self.intersect(other)
        if not intersection:
            # Do not intersect
            return 0

        intersection_area = intersection.area()

        # area of the union
        union_area = self.area() + other.area() - intersection_area

        # intersection over union
        return intersection_area / union_area


def _track_faces(frames, min_score=0.5):
    tracks = []

    for frame, faces in enumerate(frames):
        for face in faces:
            # The stats
            best_track_index = -1
            best_track_score = 0

            # Check if this face matches a track
            for track_index, track in enumerate(tracks):
                # Compare against the most recent instance of the track
                score = face.intersection_over_union(track[-1])
                if score > best_track_score:
                    best_track_score = score
                    best_track_index = track_index

            # Did we find a track?
            if best_track_score >= min_score:
                track = tracks[best_track_index]
                track.append(face)
            else:
                # New track
                tracks.append([face])

    return tracks


def _interpolate(a, b, t):
    return a + (b - a) * t


def _interpolate_boxes(box1, box2, t):
    return Box(
        int(_interpolate(box1.top, box2.top, t)),
        int(_interpolate(box1.right, box2.right, t)),
        int(_interpolate(box1.bottom, box2.bottom, t)),
        int(_interpolate(box1.left, box2.left, t))
    )


def _interpolate_faces(frames, tracking_frame_distance, tracking_confidence):
    tracks = _track_faces(frames, tracking_confidence)

    previous_faces = [
        (-1, track[0]) for track in tracks
    ]

    for frame, faces_in_frame in enumerate(frames):
        for face in faces_in_frame:
            # which track?
            track_index = -1
            for index, track in enumerate(tracks):
                if face in track:
                    track_index = index
                    break

            if track_index < 0:
                raise Exception(f"Could not find track for face {face}")

            # When was it last shown?
            previous_frame, previous_face = previous_faces[track_index]
            frame_distance = frame - previous_frame
            if 1 < frame_distance < tracking_frame_distance:
                frames_to_interpolate = frame_distance - 1
                # interpolate back
                for offset, dt in enumerate(np.linspace(0, 1, frames_to_interpolate + 2)[1:-1]):
                    new_face = _interpolate_boxes(previous_face, face, dt)
                    frame_to_fix = frames[previous_frame+1+offset]
                    frame_to_fix.append(new_face)

            previous_faces[track_index] = (frame, face)

    return frames


def identify_faces_from_video(container: InputContainer,
                              detection_confidence=0.5,
                              merge_confidence=0.25,
                              tracking_frame_distance=30,
                              tracking_confidence=0.05,
                              image_size=IDENTIFY_IMAGE_SIZE,
                              progress=tqdm.tqdm,
                              stop: TerminatingCookie = None):

    faces = {stream.index: [] for stream in container.streams if stream.type == "video"}

    with FaceDetection(model_selection=0, min_detection_confidence=detection_confidence) as detection_close:
        with FaceDetection(model_selection=1, min_detection_confidence=detection_confidence) as detection_far:
            with progress(desc="Detecting faces", total=container.video.frames, unit=" frames", leave=False) as progress:
                for packet in container.demux():
                    if packet.stream.type == "video":
                        for frame in packet.decode():
                            if stop:
                                stop.throwIfTerminated()

                            image = frame.to_image()
                            detected_faces = _identify_faces_from_image(
                                image, detection_close, detection_far, merge_confidence, image_size)

                            faces[packet.stream.index].append(detected_faces)

                            if packet.stream == container.video:
                                progress.update()

    # Convert the coords to something meaningful
    return {index: _interpolate_faces(faces, tracking_frame_distance, tracking_confidence) for index, faces in faces.items()}


def identify_faces_from_image(image: Image,
                              detection_confidence=0.5,
                              merge_confidence=0.25,
                              image_size=IDENTIFY_IMAGE_SIZE):

    with FaceDetection(model_selection=0, min_detection_confidence=detection_confidence) as detection_close:
        with FaceDetection(model_selection=1, min_detection_confidence=detection_confidence) as detection_far:
            return _identify_faces_from_image(image, detection_close, detection_far, merge_confidence, image_size)



# Copyright (C) 2025, Simona Dimitrova

import os
import threading
import wx

from faceblur.app import get_supported_filenames
from faceblur.app import faceblur
from faceblur.threading import TerminatingCookie


class Drop(wx.FileDropTarget):
    def __init__(self, window):
        super().__init__()
        self._window = window

    def OnDropFiles(self, x, y, filenames):
        def on_error(message):
            wx.MessageDialog(None, message, "Warning", wx.OK | wx.CENTER | wx.ICON_WARNING).ShowModal()
        filenames = get_supported_filenames(filenames, on_error)

        for filename in filenames:
            filename = os.path.abspath(filename)

            # Add only if not added by the user before
            if filename not in self._window._file_list.GetItems():
                self._window._file_list.Append(filename)

        return True


DEFAULT_STRENGTH = 1.0
DEFAULT_CONFIDENCE = 0.5


class MainWindow(wx.Frame):
    def __init__(self, parent, title):
        super().__init__(parent, title=title, size=(600, 400))

        self._thread = None
        self._cookie = None

        # Main panel and sizer
        panel = wx.Panel(self)
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)

        # List of files on the left
        self._file_list = wx.ListBox(panel, style=wx.LB_EXTENDED)
        self._file_list.Bind(wx.EVT_KEY_DOWN, self._list_on_key_down)
        main_sizer.Add(self._file_list, 1, wx.EXPAND | wx.ALL, 5)

        # Right panel
        right_panel = wx.Panel(panel)
        right_sizer = wx.BoxSizer(wx.VERTICAL)

        # "Options" Panel with number inputs
        options_panel = wx.StaticBox(right_panel, label="Options")
        options_sizer = wx.StaticBoxSizer(options_panel, wx.VERTICAL)

        self._strength = wx.SpinCtrlDouble(right_panel, value=str(DEFAULT_STRENGTH), min=0, max=10, inc=0.1)
        options_sizer.Add(wx.StaticText(right_panel, label="Blur strength"), 0, wx.LEFT | wx.TOP, 5)
        options_sizer.Add(self._strength, 0, wx.EXPAND | wx.ALL, 5)

        self._confidence = wx.SpinCtrlDouble(right_panel, value=str(DEFAULT_CONFIDENCE), min=0, max=1, inc=0.01)
        options_sizer.Add(wx.StaticText(right_panel, label="Detection confidence"), 0, wx.LEFT | wx.TOP, 5)
        options_sizer.Add(self._confidence, 0, wx.EXPAND | wx.ALL, 5)

        right_sizer.Add(options_sizer, 0, wx.EXPAND | wx.ALL, 5)

        # Button(s) on the right
        button_panel = wx.Panel(right_panel)
        button_sizer = wx.BoxSizer(wx.VERTICAL)

        self._reset_button = wx.Button(button_panel, label="Reset")
        self._reset_button.Bind(wx.EVT_BUTTON, self._on_reset)

        self._start_button = wx.Button(button_panel, label="Start")
        self._start_button.Bind(wx.EVT_BUTTON, self._on_start)
        self._start_button.SetDefault()

        self._stop_button = wx.Button(button_panel, label="Stop")
        self._stop_button.Bind(wx.EVT_BUTTON, self._on_stop)
        self._stop_button.Enable(False)

        self._buttons = [
            self._reset_button,
            self._start_button,
            self._stop_button,
        ]

        for button in self._buttons:
            button_sizer.Add(button, 0, wx.EXPAND | wx.ALL, 5)

        button_panel.SetSizer(button_sizer)
        right_sizer.Add(button_panel, 0, wx.EXPAND | wx.ALL, 5)

        right_panel.SetSizer(right_sizer)
        main_sizer.Add(right_panel, 0, wx.EXPAND | wx.ALL, 5)

        # Set the main panel sizer
        panel.SetSizer(main_sizer)

        # Add a status bar with progress bars
        self._status_bar = wx.StatusBar(self)
        self.SetStatusBar(self._status_bar)

        # Create 3 fields in the status bar
        self._status_bar.SetFieldsCount(3)
        self._status_bar.SetStatusWidths([-1, 150, 150])  # -1 = stretch, others fixed

        # Add progress bars to the last two fields
        self._progress_bar1 = wx.Gauge(self._status_bar)
        self._progress_bar2 = wx.Gauge(self._status_bar)

        # Position progress bars within the status bar
        self.Bind(wx.EVT_SIZE, self.on_resize_statusbar)  # Keep layout consistent during resizing
        self.position_progress_bars()

        # Simulate some progress changes
        self._progress_bar1.SetValue(60)
        self._progress_bar2.SetValue(30)

        # Support drag & drop
        self.SetDropTarget(Drop(self))

        # Show the window
        self.Centre()
        self.Show()

    def position_progress_bars(self):
        # Dynamically position progress bars based on the status bar's dimensions
        rect1 = self._status_bar.GetFieldRect(1)
        rect2 = self._status_bar.GetFieldRect(2)

        self._progress_bar1.SetSize(rect1)
        self._progress_bar1.SetPosition((rect1.x, rect1.y))
        self._progress_bar1.SetValue(0)

        self._progress_bar2.SetPosition((rect2.x + 5, rect2.y + (rect2.height - self._progress_bar2.GetSize()[1]) // 2))

    def on_resize_statusbar(self, event):
        self.position_progress_bars()
        event.Skip()

    def _list_on_key_down(self, event):
        # Check for Ctrl+A (Select All)
        if event.GetKeyCode() == ord('A') and event.ControlDown():
            # Select all items (one by one)
            for index in range(self._file_list.GetCount()):
                self._file_list.SetSelection(index)

        # Check if the Delete key is pressed
        elif event.GetKeyCode() == wx.WXK_DELETE:
            # Get a list of selected indices
            selections = self._file_list.GetSelections()
            if selections:
                # Reverse the selection order to avoid index shifting issues
                for index in reversed(selections):
                    self._file_list.Delete(index)
        else:
            # Pass other key events to the list box
            event.Skip()

    def _on_reset(self, event):
        self._strength.SetValue(DEFAULT_STRENGTH)
        self._confidence.SetValue(DEFAULT_CONFIDENCE)

    def _remove_file(self, filename):
        for index, f in enumerate(self._file_list.GetItems()):
            if f == filename:
                self._file_list.Delete(index)

                # Assumes no duplicates in the list
                break

    def _set_enablements(self, stopped):
        self._file_list.Enable(stopped)
        self._strength.Enable(stopped)
        self._confidence.Enable(stopped)
        self._reset_button.Enable(stopped)
        self._start_button.Enable(stopped)
        self._stop_button.Enable(not stopped)

    def _thread_done(self):
        assert self._thread

        self._thread.join()
        self._cookie = None
        self._thread = None

        self._set_enablements(True)

    def _on_done(self, filename):
        if filename:
            # 1 file has finished. Remove it from the list
            wx.CallAfter(self._remove_file, filename)
        else:
            # All files have finished
            wx.CallAfter(self._thread_done)

    def _on_start(self, event):
        assert not self._thread
        assert not self._cookie

        if not self._file_list.GetCount():
            # Nothing to do
            return

        with wx.DirDialog(None, "Choose output folder", style=wx.DD_DEFAULT_STYLE) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                output = dlg.GetPath()
            else:
                # Did not select
                return

        self._cookie = TerminatingCookie()

        kwargs = {
            "inputs": self._file_list.GetItems(),
            "output": output,
            "strength": self._strength.GetValue(),
            "confidence": self._confidence.GetValue(),
            # "progress_type": None,
            "on_done": self._on_done,
            "stop": self._cookie,
        }

        self._set_enablements(False)

        self._thread = threading.Thread(target=faceblur, kwargs=kwargs)
        self._thread.start()

    def _on_stop(self, event):
        assert self._cookie
        self._cookie.requestTermination()


def main():
    app = wx.App(False)
    frame = MainWindow(None, "FaceBlur: Automatic Photo and Video Deidentifier")
    app.MainLoop()


if __name__ == "__main__":
    main()
